﻿Public Class Form1
    Dim r As IO.BinaryReader

    Dim pack As Package

    Function InputFileName() As String
        'Return "D:\!2014\unreal beta database\sources\Full_0.82_1996\CLASSES\root.ucx"
        Return OpenDlg.FileName
    End Function

    Sub OpenFile()
        If r IsNot Nothing Then
            r.Close()
        End If

        r = New IO.BinaryReader(New IO.FileStream(InputFileName(), IO.FileMode.Open, IO.FileAccess.Read))
    End Sub

    Sub ReadFile()
        pack = New Package

        'go to beginning of file header
        pack.FileHeaderPtr = r.BaseStream.Length - 164
        r.BaseStream.Position = pack.FileHeaderPtr

        'read file header
        pack.PackageVersion = r.ReadUInt32()

        pack.ExportCount = r.ReadUInt32()

        r.BaseStream.Position = pack.FileHeaderPtr + 24
        pack.ExportTablePtr = r.ReadUInt32()
        pack.ResHeadersPtr = r.ReadUInt32()

        'read export table

        lvExportTable.Items.Clear()
        For i = 0 To pack.ExportCount - 1
            r.BaseStream.Position = pack.ExportTablePtr + 44 * i
            Dim curResName As String = ReadString(r, 36)
            r.BaseStream.Position = pack.ExportTablePtr + 44 * i + 36
            Dim curResType As UInt32 = r.ReadUInt32
            Dim curResHeaderSize As UInt32 = r.ReadUInt32

            pack.ExportTable.Add(New ExportItem With {.Name = curResName, .TypeNum = curResType, .HeaderSize = curResHeaderSize})

        Next

        'calculate header pointers
        Dim sum As Integer = pack.ResHeadersPtr
        For Each i In pack.ExportTable
            i.HeaderPtr = sum
            sum += i.HeaderSize
        Next

        'read data pointers & lengths
        For Each i In pack.ExportTable
            r.BaseStream.Position = i.HeaderPtr + 60
            i.DataPtr = r.ReadUInt32()
            i.DataSize = r.ReadUInt32()
        Next



        'fill GUI lists

        For Each i In pack.ExportTable
            Dim item = lvExportTable.Items.Add(i.Name)
            item.SubItems.AddRange({TypeToName(i.TypeNum), i.HeaderSize, i.HeaderPtr, i.DataPtr, i.DataSize})
        Next

    End Sub

    Function TypeToName(typeID As Integer) As String
        Dim typeName As String

        Select Case typeID
            Case 3
                typeName = "brush list"
            Case 4
                typeName = "code"
            Case 5
                typeName = "texture"
            Case 7
                typeName = "palette"
            Case 10
                typeName = "class"
            Case 11
                typeName = "actor list"
            Case 12
                typeName = "sound"
            Case 14
                typeName = "mesh"
            Case 20
                typeName = "model"
            Case 21
                typeName = "brush"
            Case 35
                typeName = "enum"
            Case Else
                Return typeID
        End Select

        Return typeName & "/" & typeID
    End Function

    Function ReadString(ByVal r As IO.BinaryReader, Optional ByVal maxLen As UInteger = 0) As String
        Dim str As New System.Text.StringBuilder
        Dim len As UInteger = 0
        Dim curchr As Byte

        Do
            curchr = r.ReadByte
            len += 1
            If curchr = 0 Then
                Exit Do
            End If
            str.Append(Chr(curchr))
        Loop Until (len >= maxLen And maxLen > 0)

        Return str.ToString
    End Function

    Function ReadTexture(resNum As UInteger) As Bitmap
        Dim resItem As ExportItem = pack.ExportTable(resNum)
        Dim headerPtr As UInt32 = resItem.HeaderPtr
        Dim width, height, paletteNum As Int32
        r.BaseStream.Position = headerPtr + 80
        paletteNum = r.ReadUInt32() - 1 '1-based to 0-based
        r.BaseStream.Position = headerPtr + 160
        width = r.ReadUInt32()
        height = r.ReadUInt32()

        'read palette
        Dim palette(256 - 1) As Color
        r.BaseStream.Position = pack.ExportTable(paletteNum).DataPtr
        For i = 0 To 255
            palette(i) = RGBXToColor(r.ReadInt32)
        Next

        'go to beginning of largest (last) mipmap
        r.BaseStream.Position = resItem.DataPtr + resItem.DataSize - (width * height)
        Dim bmp As New Bitmap(width, height)
        Dim fastBmp As New FastPixel(bmp)
        fastBmp.Lock()

        For y = 0 To height - 1
            For x = 0 To width - 1
                fastBmp.SetPixel(x, y, palette(r.ReadByte()))
            Next
        Next

        fastBmp.Unlock(True)

        Return bmp

    End Function

    Function RGBXToColor(num As Integer) As Color
        Return Color.FromArgb(num And 255, (num >> 8) And 255, (num >> 16) And 255)
    End Function

    Private Sub btnCopyDataAddr_Click(sender As System.Object, e As System.EventArgs) Handles btnCopyDataAddr.Click
        If lvExportTable.SelectedItems.Count = 1 Then
            My.Computer.Clipboard.SetText(lvExportTable.SelectedItems(0).SubItems(4).Text)
        End If

    End Sub

    Private Sub btnCopyHeaderAddr_Click(sender As System.Object, e As System.EventArgs) Handles btnCopyHeaderAddr.Click
        If lvExportTable.SelectedItems.Count = 1 Then
            My.Computer.Clipboard.SetText(lvExportTable.SelectedItems(0).SubItems(3).Text)
        End If
    End Sub

    Private Sub btnPreview_Click(sender As System.Object, e As System.EventArgs) Handles btnPreview.Click
        If lvExportTable.SelectedItems.Count <> 1 Then Return

        Dim selectedItem As ExportItem = pack.ExportTable(lvExportTable.SelectedIndices(0))

        Select Case selectedItem.TypeNum
            Case 4 'code
                r.BaseStream.Position = selectedItem.DataPtr
                MessageBox.Show(ReadString(r, selectedItem.DataSize))
            Case 12 'sound
                r.BaseStream.Position = selectedItem.DataPtr + 192
                Dim wavSize As UInt32 = r.ReadUInt32()
                r.BaseStream.Position = selectedItem.DataPtr + 252
                My.Computer.Audio.Play(r.ReadBytes(wavSize), AudioPlayMode.Background)
            Case 5 'texture
                r.BaseStream.Position = selectedItem.HeaderPtr + 80
                Dim paletteNum As UInt32 = r.ReadUInt32()
                Dim width, height As UInt32
                r.BaseStream.Position = selectedItem.HeaderPtr + 160
                width = r.ReadUInt32()
                height = r.ReadUInt32()
                r.BaseStream.Position = selectedItem.HeaderPtr + 170
                Dim hasMipmaps As Boolean = r.ReadByte()

                MessageBox.Show("PaletteNum (starts from 1): " & paletteNum & vbCrLf &
                                "Resolution: " & width & "x" & height & vbCrLf &
                                "Has mipmaps: " & hasMipmaps)

                TexPreviewWindow.Show()
                TexPreviewWindow.PictureBox1.Image = ReadTexture(lvExportTable.SelectedIndices(0))

            Case Else
                MessageBox.Show("Type not supported for preview.")
        End Select
    End Sub

    Private Sub lvExportTable_SelectedIndexChanged(sender As System.Object, e As System.EventArgs) Handles lvExportTable.SelectedIndexChanged
        If lvExportTable.SelectedItems.Count = 1 Then
            lbResourceNum.Text = lvExportTable.SelectedIndices(0)
        Else
            lbResourceNum.Text = ""
        End If

    End Sub

    Private Sub Form1_Load(sender As System.Object, e As System.EventArgs) Handles MyBase.Load
        
    End Sub

    Private Sub btnSaveBinData_Click(sender As System.Object, e As System.EventArgs) Handles btnSaveBinData.Click
        If lvExportTable.SelectedItems.Count <> 1 Then Return
        Dim selectedItem As ExportItem = pack.ExportTable(lvExportTable.SelectedIndices(0))

        SaveBinDlg.FileName = selectedItem.Name
        If SaveBinDlg.ShowDialog = Windows.Forms.DialogResult.OK Then
            r.BaseStream.Position = selectedItem.DataPtr
            IO.File.WriteAllBytes(SaveBinDlg.FileName, r.ReadBytes(selectedItem.DataSize))
        End If
    End Sub

    Private Sub btnSaveBinHeader_Click(sender As System.Object, e As System.EventArgs) Handles btnSaveBinHeader.Click
        If lvExportTable.SelectedItems.Count <> 1 Then Return
        Dim selectedItem As ExportItem = pack.ExportTable(lvExportTable.SelectedIndices(0))

        SaveBinDlg.FileName = selectedItem.Name & " header"
        If SaveBinDlg.ShowDialog = Windows.Forms.DialogResult.OK Then
            r.BaseStream.Position = selectedItem.HeaderPtr
            IO.File.WriteAllBytes(SaveBinDlg.FileName, r.ReadBytes(selectedItem.HeaderSize))
        End If
    End Sub

    Private Sub btnExport_Click(sender As System.Object, e As System.EventArgs) Handles btnExport.Click
        If FolderDlg.ShowDialog = Windows.Forms.DialogResult.OK Then
            For i = 0 To pack.ExportTable.Count - 1
                Dim item As ExportItem = pack.ExportTable(i)

                If Exportable(i) Then
                    Try
                        ExportResource(i, IO.Path.Combine(FolderDlg.SelectedPath, item.Name & ExportResourceExt(i)))
                    Catch ex As Exception
                        MessageBox.Show("Exporting " & i & " (" & pack.ExportTable(i).Name & ") failed." & vbCrLf & vbCrLf & "Error type:" & vbCrLf & ex.GetType.Name & vbCrLf & "Details:" & ex.Message & vbCrLf & "Stack:" & ex.StackTrace, "Export error", MessageBoxButtons.OK, MessageBoxIcon.Error)
                    End Try
                End If

                'Select Case item.TypeNum
                '    Case 4 'code
                '        r.BaseStream.Position = item.DataPtr
                '        IO.File.WriteAllBytes(IO.Path.Combine(FolderDlg.SelectedPath, item.Name & ".txt"), r.ReadBytes(item.DataSize))
                '    Case 5 'texture
                '        ReadTexture(i).Save(IO.Path.Combine(FolderDlg.SelectedPath, item.Name & ".png"), Imaging.ImageFormat.Png)
                '    Case 12 'sound
                '        r.BaseStream.Position = item.DataPtr + 252
                '        IO.File.WriteAllBytes(IO.Path.Combine(FolderDlg.SelectedPath, item.Name & ".wav"), r.ReadBytes(item.DataSize - 252))
                'End Select

                ProgressBar1.Value = i / pack.ExportTable.Count * 100
                ProgressBar1.Refresh()
            Next

            MessageBox.Show("Finished.")
        End If
    End Sub

    ''' <summary>
    ''' Tries to export resource [num]. Warning: in case of mesh resources, two files will be produced, with filename endings "_d" and "_a".
    ''' </summary>
    ''' <returns>Returns whether exporting was possible.</returns>
    Function ExportResource(num As Integer, file As String) As Boolean
        Dim res As ExportItem = pack.ExportTable(num)

        Select Case res.TypeNum
            Case 4 'code
                r.BaseStream.Position = res.DataPtr
                IO.File.WriteAllBytes(file, r.ReadBytes(res.DataSize))
                Return True
            Case 5 'texture
                ReadTexture(num).Save(file, Imaging.ImageFormat.Png)
                Return True
            Case 12 'sound
                r.BaseStream.Position = res.DataPtr + 252
                IO.File.WriteAllBytes(file, r.ReadBytes(res.DataSize - 252))
                Return True
            Case 14 'mesh
                r.BaseStream.Position = res.HeaderPtr + 152
                Dim polygonCount As UInt16 = r.ReadUInt16()
                r.BaseStream.Position = res.HeaderPtr + 160
                Dim vertexCount As UInt16 = r.ReadUInt16()
                r.BaseStream.Position = res.HeaderPtr + 176
                Dim frameCount As UInt16 = r.ReadUInt16()

                Dim dataFileName As String
                Dim animFileName As String
                Dim fileNameStart As String
                Dim fileNameEnd As String
                fileNameStart = IO.Path.Combine(IO.Path.GetDirectoryName(file), IO.Path.GetFileNameWithoutExtension(file))
                fileNameEnd = IO.Path.GetExtension(file)
                dataFileName = fileNameStart & "_d" & fileNameEnd
                animFileName = fileNameStart & "_a" & fileNameEnd

                Dim dataWriter As New IO.BinaryWriter(New IO.FileStream(dataFileName, IO.FileMode.CreateNew, IO.FileAccess.ReadWrite))
                dataWriter.Write(CUShort(polygonCount))
                dataWriter.Write(CUShort(vertexCount))
                dataWriter.Seek(44, IO.SeekOrigin.Current)
                r.BaseStream.Position = res.DataPtr
                dataWriter.Write(r.ReadBytes(16 * polygonCount))
                dataWriter.Close()

                Dim animWriter As New IO.BinaryWriter(New IO.FileStream(animFileName, IO.FileMode.CreateNew, IO.FileAccess.ReadWrite))
                animWriter.Write(CUShort(frameCount))
                animWriter.Write(CUShort(vertexCount * 4))
                animWriter.Write(r.ReadBytes(CInt(frameCount) * CInt(vertexCount) * 4))
                animWriter.Close()
                Return True
            Case Else
                Return False
        End Select
    End Function

    ''' <summary>
    ''' Returns file extension for exported resource [num]. Returns "" if not exportable.
    ''' </summary>
    Function ExportResourceExt(num As Integer) As String
        Dim res As ExportItem = pack.ExportTable(num)

        Select Case res.TypeNum
            Case 4
                Return ".txt"
            Case 5
                Return ".png"
            Case 12
                Return ".wav"
            Case 14
                Return ".3d"
        End Select

        Return ""
    End Function

    ''' <summary>
    ''' Returns whether resource [num] can be exported.
    ''' </summary>
    Function Exportable(num As Integer) As Boolean
        Select Case pack.ExportTable(num).TypeNum
            Case 4, 5, 12, 14
                Return True
            Case Else
                Return False
        End Select
    End Function

    Private Sub btnOpen_Click(sender As System.Object, e As System.EventArgs) Handles btnOpen.Click
        If OpenDlg.ShowDialog = Windows.Forms.DialogResult.OK Then
            OpenFile()
            ReadFile()
        End If
        
    End Sub
End Class
